#pragma clang diagnostic push
#pragma ide diagnostic ignored "performance-unnecessary-value-param"
#pragma ide diagnostic ignored "readability-convert-member-functions-to-static"

#include <filesystem>
#include "controllers.h"
#include "dto/dto.h"

char windows_invalid_chars[] = "\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1A\x1B\x1C\x1D\x1E\x1F<>:\"/\\|";

std::string generate_path(db::INode node) {
    db::MapperInode inode_mapper(drogon::app().getDbClient());
    std::stack<db::INode> path;
    path.push(node);
    while (node.getParentId() != nullptr) {
        node = inode_mapper.findByPrimaryKey(node.getValueOfParentId());
        path.push(node);
    }
    std::stringstream ss;
    while (!path.empty()) {
        const db::INode& seg = path.top();
        ss << seg.getValueOfName();
        if (seg.getValueOfIsFile() == 0) ss << '/';
        path.pop();
    }
    return ss.str();
}

namespace api {
    std::optional<db::INode> fs::get_node(uint64_t node) {
        db::MapperInode inode_mapper(drogon::app().getDbClient());
        try {
            return inode_mapper.findByPrimaryKey(node);
        } catch (const std::exception&) {
            return std::nullopt;
        }
    }

    std::optional<db::INode> fs::get_node_and_validate(const db::User &user, uint64_t node) {
        auto inode = get_node(node);
        if (!inode.has_value()) return std::nullopt;
        if (inode->getValueOfOwnerId() != user.getValueOfId()) return std::nullopt;
        return inode;
    }

    std::vector<db::INode> fs::get_children(const db::INode& parent) {
        db::MapperInode inode_mapper(drogon::app().getDbClient());
        return inode_mapper.findBy(db::Criteria(db::INode::Cols::_parent_id, db::CompareOps::EQ, parent.getValueOfId()));
    }

    std::variant<db::INode, std::string> fs::create_node(std::string name, const db::User& owner, bool file, const std::optional<uint64_t> &parent, bool force) {
        // Stolen from https://github.com/boostorg/filesystem/blob/develop/src/portability.cpp
        if (!force)
            if (name.empty() || name[0] == ' ' || name.find_first_of(windows_invalid_chars, 0, sizeof(windows_invalid_chars)) != std::string::npos || *(name.end() - 1) == ' ' || *(name.end() - 1) == '.' || name == "." || name == "..")
                return {"Invalid name"};

        db::INode node;
        node.setIsFile(file ? 1 : 0);
        node.setName(name);
        node.setOwnerId(owner.getValueOfId());
        if (parent.has_value()) {
            auto parent_node =  get_node_and_validate(owner, *parent);
            if (!parent_node.has_value())
                return {"Invalid parent"};
            if (parent_node->getValueOfIsFile() != 0)
                return {"Can't use file as parent"};
            auto children = get_children(*parent_node);
            for (const auto& child : children)
                if (child.getValueOfName() == name)
                    return {"File/Folder already exists"};
            node.setParentId(*parent);
        }
        db::MapperInode inode_mapper(drogon::app().getDbClient());
        inode_mapper.insert(node);
        return {node};
    }

    void fs::delete_node(db::INode node, bool allow_root) {
        if (node.getValueOfParentId() == 0 && (!allow_root)) return;
        if (node.getValueOfIsFile() == 0) {
            auto children =  get_children(node);
            for (const auto& child : children) delete_node(child, false);
        } else {
            std::filesystem::path p("./files");
            p /= std::to_string(node.getValueOfId());
            std::filesystem::remove(p);
        }
        db::MapperInode inode_mapper(drogon::app().getDbClient());
        inode_mapper.deleteOne(node);
    }

    void fs::root(req_type req, cbk_type cbk) {
        db::User user = dto::get_user(req);
        cbk(dto::Responses::get_root_res(user.getValueOfRootId()));
    }

    void fs::node(req_type req, cbk_type cbk, uint64_t node) {
        db::User user = dto::get_user(req);
        auto inode =  get_node_and_validate(user, node);
        if (!inode.has_value())
            cbk(dto::Responses::get_badreq_res("Unknown node"));
        else if (inode->getValueOfIsFile() == 0) {
            std::vector<uint64_t> children;
            for (const db::INode& child : get_children(*inode)) children.push_back(child.getValueOfId());
            cbk(dto::Responses::get_node_folder_res(
                    inode->getValueOfId(),
                    inode->getValueOfName(),
                    inode->getParentId(),
                    children
            ));
        } else
            cbk(dto::Responses::get_node_file_res(
                    inode->getValueOfId(),
                    inode->getValueOfName(),
                    inode->getParentId(),
                    inode->getValueOfSize()
            ));
    }

    void fs::path(req_type req, cbk_type cbk, uint64_t node) {
        db::User user = dto::get_user(req);
        auto inode = get_node_and_validate(user, node);
        if (!inode.has_value())
            cbk(dto::Responses::get_badreq_res("Unknown node"));
        else
            cbk(dto::Responses::get_path_res( generate_path(*inode)));
    }

    template<bool file>
    void fs::create_node_req(req_type req, cbk_type cbk) {
        db::User user = dto::get_user(req);
        Json::Value& json = *req->jsonObject();
        try {
            uint64_t parent = dto::json_get<uint64_t>(json, "parent").value();
            std::string name = dto::json_get<std::string>(json, "name").value();

            auto new_node = create_node(name, user, file, std::make_optional(parent));
            if (std::holds_alternative<std::string>(new_node))
                cbk(dto::Responses::get_badreq_res(std::get<std::string>(new_node)));
            else
                cbk(dto::Responses::get_new_node_res(std::get<db::INode>(new_node).getValueOfId()));
        } catch (const std::exception&) {
            cbk(dto::Responses::get_badreq_res("Validation error"));
        }
    }

    void fs::delete_node_req(req_type req, cbk_type cbk, uint64_t node) {
        db::User user = dto::get_user(req);
        auto inode = get_node_and_validate(user, node);
        if (!inode.has_value())
            cbk(dto::Responses::get_badreq_res("Unknown node"));
        else if (inode->getValueOfParentId() == 0)
            cbk(dto::Responses::get_badreq_res("Can't delete root"));
        else {
             delete_node(*inode);
            cbk(dto::Responses::get_success_res());
        }
    }

    void fs::upload(req_type req, cbk_type cbk, uint64_t node) {
        db::User user = dto::get_user(req);

        auto inode = get_node_and_validate(user, node);
        if (!inode.has_value())
            return cbk(dto::Responses::get_badreq_res("Unknown node"));
        if (inode->getValueOfIsFile() == 0)
            return cbk(dto::Responses::get_badreq_res("Can't upload to a directory"));

        drogon::MultiPartParser mpp;
        if (mpp.parse(req) != 0)
            return cbk(dto::Responses::get_badreq_res("Failed to parse files"));
        if (mpp.getFiles().size() != 1)
            return cbk(dto::Responses::get_badreq_res("Exactly 1 file needed"));

        const drogon::HttpFile& file = mpp.getFiles().at(0);

        std::filesystem::path p("./files");
        p /= std::to_string(inode->getValueOfId());

        file.saveAs(p.string());

        inode->setSize(file.fileLength());
        db::MapperInode inode_mapper(drogon::app().getDbClient());
        inode_mapper.update(*inode);
        cbk(dto::Responses::get_success_res());
    }

    void fs::download(req_type req, cbk_type cbk) {
        db::User user = dto::get_user(req);

        auto node_id = req->getOptionalParameter<uint64_t>("id");
        if (!node_id.has_value()) {
            cbk(dto::Responses::get_badreq_res("Invalid node"));
            return;
        }
        auto inode =  get_node_and_validate(user, *node_id);
        if (!inode.has_value()) {
            cbk(dto::Responses::get_badreq_res("Invalid node"));
            return;
        }

        std::filesystem::path p("./files");
        p /= std::to_string(inode->getValueOfId());

        cbk(drogon::HttpResponse::newFileResponse(
            p.string(),
            inode->getValueOfName()
        ));
    }
}
#pragma clang diagnostic pop